/* l_linux.c -- stub loader for Linux

   This file is part of the UPX executable compressor.

   Copyright (C) 1996-1999 Markus Franz Xaver Johannes Oberhumer
   Copyright (C) 1996-1999 Laszlo Molnar

   This file is part of the UPX stub used in compressed
   linux/i386 executables. All Rights Reserved.

   Markus F.X.J. Oberhumer                   Laszlo Molnar
   markus.oberhumer@jk.uni-linz.ac.at        ml1050@cdata.tvnet.hu
 */


#if !defined(__linux__) || !defined(__i386__)
#  error "this stub must be compiled under linux/i386"
#endif

#define __need_timeval
#include <sys/types.h>
#include <fcntl.h>
#include <sched.h>
#include <time.h>
#include <linux/errno.h>
#include <linux/mman.h>
#include <linux/unistd.h>


/*************************************************************************
// configuration section
**************************************************************************/

// use fork() instead of /proc/self/fd/X
//#define USE_FORK

// use malloc instead of the bss segement
#define USE_MALLOC


/*************************************************************************
// syscalls
//
// Because of different <asm/unistd.h> versions and subtle bugs
// in both gcc and egcs we define all syscalls manually.
// Also, errno conversion is not necessary in our case.
**************************************************************************/

#undef _syscall0
#undef _syscall1
#undef _syscall2
#undef _syscall3
#ifndef __NR__exit
#  define __NR__exit __NR_exit
#endif

#if defined(__i386__)
#define _syscall0(type,name) \
type name(void) \
{ \
long __res; \
__asm__ volatile ("int $0x80" \
        : "=a" (__res) \
        : "a" (__NR_##name)); \
return (type) __res; \
}

#define _syscall1(type,name,type1,arg1) \
type name(type1 arg1) \
{ \
long __res; \
__asm__ volatile ("int $0x80" \
        : "=a" (__res) \
        : "a" (__NR_##name),"b" ((long)(arg1))); \
return (type) __res; \
}

#define _syscall2(type,name,type1,arg1,type2,arg2) \
type name(type1 arg1,type2 arg2) \
{ \
long __res; \
__asm__ volatile ("int $0x80" \
        : "=a" (__res) \
        : "a" (__NR_##name),"b" ((long)(arg1)),"c" ((long)(arg2))); \
return (type) __res; \
}

#define _syscall3(type,name,type1,arg1,type2,arg2,type3,arg3) \
type name(type1 arg1,type2 arg2,type3 arg3) \
{ \
long __res; \
__asm__ volatile ("int $0x80" \
        : "=a" (__res) \
        : "a" (__NR_##name),"b" ((long)(arg1)),"c" ((long)(arg2)), \
                  "d" ((long)(arg3))); \
return (type) __res; \
}
#endif /* __i386__ */

#define fcntl           syscall_fcntl
#define open            syscall_open
#define nanosleep       syscall_nanosleep

static inline _syscall1(int,close,int,fd)
static inline _syscall3(int,execve,const char *,file,char **,argv,char **,envp)
static inline _syscall1(int,_exit,int,exitcode)
static inline _syscall3(int,fcntl,int,fd,int,cmd,long,arg)
static inline _syscall2(int,ftruncate,int,fd,size_t,len)
static inline _syscall0(pid_t,fork)
static inline _syscall0(pid_t,getpid)
static inline _syscall2(int,gettimeofday,struct timeval *,tv,void *,tz)
static inline _syscall3(off_t,lseek,int,fd,off_t,offset,int,count)
static inline _syscall1(caddr_t,mmap,int *,args)
static inline _syscall3(int,msync,const void *,start,size_t,length,int,flags)
static inline _syscall2(int,munmap,void *,start,size_t,length)
static inline _syscall2(int,nanosleep,const struct timespec *,rqtp,struct timespec *,rmtp)
static inline _syscall3(int,open,const char *,file,int,flag,int,mode)
static inline _syscall3(int,read,int,fd,char *,buf,off_t,count)
static inline _syscall3(pid_t,waitpid,pid_t,pid,int *,wait_stat,int,options)
static inline _syscall3(int,write,int,fd,const char *,buf,off_t,count)
static inline _syscall1(int,unlink,const char *,file)
#define exit _exit


#undef int32_t
#undef uint32_t
#define int32_t         int
#define uint32_t        unsigned int


/*************************************************************************
// util
**************************************************************************/

#undef xread
#undef xwrite

#if 1
static int xread(int fd, void *buf, int count) __attribute__((stdcall));
static int xread(int fd, void *buf, int count)
{
    while (count > 0)
    {
        int n = read(fd, buf, count);
        if (n == -EINTR)
            continue;
        if (n <= 0)
            break;
        buf += n;               // gcc extension: add to void *
        count -= n;
    }
    return count;
}
#else
#define xread(fd,buf,count)     ((count) - read(fd,buf,count))
#endif


#if 1
static __inline__ int xwrite(int fd, const void *buf, int count)
{
    while (count > 0)
    {
        int n = write(fd, buf, count);
        if (n == -EINTR)
            continue;
        if (n <= 0)
            break;
        buf += n;               // gcc extension: add to void *
        count -= n;
    }
    return count;
}
#else
#define xwrite(fd,buf,count)    ((count) - write(fd,buf,count))
#endif


/*************************************************************************
// UPX & NRV stuff
**************************************************************************/

// must be the same as in p_linux.cpp !
#define OVERHEAD        2048
#if !defined(USE_MALLOC)
#  define BLOCKSIZE     (512*1024)
#endif


// patch & magic constants for our loader (le32 format)
#define UPX1            0x31585055          // "UPX1"
#define UPX2            0x32585055          // "UPX2"
#define UPX_MAGIC_LE32  0x21585055          // "UPX!"


typedef int nrv_int;
typedef int nrv_int32;
typedef unsigned nrv_uint;
typedef unsigned nrv_uint32;
#define nrv_byte unsigned char
#define nrv_voidp void *

#if defined(__i386__)
#define NRV_ASM_NAME(x) __asm__("_" ## #x)
extern int
nrv2b_decompress_asm_fast ( const nrv_byte *src, nrv_uint src_len,
                            nrv_byte *dst, nrv_uint *dst_len,
                            nrv_voidp wrkmem /* NOT USED */ )
    NRV_ASM_NAME(nrv2b_decompress_asm_fast);
#define nrv2b_decompress        nrv2b_decompress_asm_fast
extern int
nrv2d_decompress_asm_fast ( const nrv_byte *src, nrv_uint src_len,
                            nrv_byte *dst, nrv_uint *dst_len,
                            nrv_voidp wrkmem /* NOT USED */ )
    NRV_ASM_NAME(nrv2d_decompress_asm_fast);
#define nrv2d_decompress        nrv2b_decompress_asm_fast
#endif /* __i386__ */


/*************************************************************************
// upx_main - called by our entry code
//
// This function is optimized for size.
**************************************************************************/

void upx_main(char *argv[], char *envp[]) __asm__("upx_main");
void upx_main(char *argv[], char *envp[])
{
    int fdi, fdo;
    uint32_t header[3];
    static char tmpname[] = "/tmp/upxAAAAAAAAAAA";  // filename max 14 chars
#if !defined(USE_FORK)
    static char tmpfd[] = "/proc/self/fd/000000000";
#endif

#if defined(USE_MALLOC)
    unsigned char *buf;
    static int malloc_args[6] = {
        0, OVERHEAD, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, -1, 0
    };
#else
    static unsigned char buf[BLOCKSIZE + OVERHEAD];
#endif


    //
    // ----- Step 1: prepare input file -----
    //

    // Open the exe.
    fdi = open("/proc/self/exe", O_RDONLY, 0);
    // No proc filesystem ? Try to open argv[0].
    if (fdi < 0)
        fdi = open(argv[0], O_RDONLY, 0);
    if (fdi < 0)
        goto error1;

    // Seek to start of compressed data.
    if (lseek(fdi, UPX1, 0) < 0)
        goto error1;
    // Paranoia. Make sure this is actually our expected executable.
    if (xread(fdi, (void *)header, 12) != 0 || header[0] != UPX2)
        goto error1;


    //
    // ----- Step 2: prepare temporary output file -----
    //

    // Compute name of temporary output file.
    // Protect against Denial of Service attacks.
    {
        char *p = tmpname + sizeof(tmpname) - 1;
        uint32_t r;

        // Compute 4 characters (~18.8 bits) from getpid().
        {
            int k = 4;
            r = (uint32_t) getpid();
            do {
                *--p += r % 26;
                r /= 26;
            } while (--k > 0);
        }

        // Provide 4 random bytes from our program id.
        r ^= header[0];
#if 0
        // Try to get 4 random bytes at runtime.
        {
            int fd = open("/dev/urandom", O_RDONLY, 0);
            if (fd < 0)
                fd = open("/dev/random", O_RDONLY, 0);
            if (fd >= 0)
            {
                (void) read(fd, (void *)&r, 4);
                (void) close(fd);
            }
        }
#elif 1
        // Try to get 4 random bytes at runtime.
        // Don't absorb precious bytes from /dev/urandom.
        {
            struct timeval tv;
            gettimeofday(&tv, 0);
            r ^= (uint32_t) tv.tv_sec;
            r ^= ((uint32_t) tv.tv_usec) << 12;     // shift into high-bits
        }
#endif
        // Compute 7 characters (32 random bits).
        {
            int k = 7;
            do {
                *--p += r % 26;
                r /= 26;
            } while (--k > 0);
        }
    }

    // Just in case, remove the file.
    {
        int err = unlink(tmpname);
        if (err != -ENOENT && err != 0)
            goto error1;
    }

    // Open the temporary output file.
    fdo = open(tmpname, O_WRONLY | O_CREAT | O_EXCL, 0700);
    if (fdo < 0)
        goto error;

    // Set expected file size.
    if (ftruncate(fdo, header[1]) != 0)
        goto error;


    //
    // ----- Step 3: setup memory -----
    //

#if defined(USE_MALLOC)
    malloc_args[1] += header[2];
    buf = mmap(malloc_args);
    if ((unsigned long) buf >= (unsigned long) -4095)
        goto error;
#else
    if (header[2] > BLOCKSIZE)
        goto error;
#endif


    //
    // ----- Step 4: decompress blocks -----
    //

    for (;;)
    {
        int32_t size[2];
        int i;

        // Read & check block sizes.
        if (xread(fdi, (void *)size, 8) != 0)
            goto error;
        if (size[0] == 0)                       // uncompressed size 0 -> EOF
        {
            if (size[1] != UPX_MAGIC_LE32)      // size[1] must be h->magic
                goto error;
            if (header[1] != 0)                 // all bytes must be written
                goto error;
            break;
        }
        if (size[0] <= 0 || size[1] <= 0)
            goto error;
        if (size[1] > size[0] || size[1] > (int32_t)header[2])
            goto error;

        // Read block.
        i = header[2] + OVERHEAD - size[1];
        if (xread(fdi, buf+i, size[1]) != 0)
            goto error;

        // Decompress block.
        if (size[1] < size[0])
        {
            // in-place decompression
            nrv_uint out_len;
            i = nrv2b_decompress(buf+i, size[1], buf, &out_len, 0);
            if (i != 0 || out_len != (nrv_uint)size[0])
                goto error;
            // i == 0 now

        }

        // Write block.
        if (xwrite(fdo, buf+i, size[0]) != 0)
            goto error;
        header[1] -= size[0];
    }


    //
    // ----- Step 5: release resources -----
    //

#if defined(USE_MALLOC)
    munmap(buf, malloc_args[1]);
#endif

    // Close files.
    if (close(fdi) != 0)
        goto error;
    if (close(fdo) != 0)
        goto error;


    //
    // ----- Step 6: start program -----
    //

#if defined(USE_FORK)

    // Fork off a subprocess to clean up.
    // We have to do this double-fork trick to keep a zombie from
    // hanging around if the spawned original program doesn't check for
    // subprocesses (as well as to prevent the real program from getting
    // confused about this subprocess it shouldn't have).
    // Thanks to Adam Ierymenko <api@one.net> for this solution.

    if (!fork())
    {
        if (!fork())
        {
            // Sleep 2 seconds, then remove the temp file.
            static const struct timespec ts = { 2, 0 };
            nanosleep(&ts, 0);
            unlink(tmpname);
        }
        exit(0);
    }

    // Wait for the first fork()'d process to die.
    waitpid(-1, (int *)0, 0);

    // Execute the original program.
    execve(tmpname, argv, envp);

#else

    // Use /proc/self/fd/X. Many thanks to Andi Kleen <ak@muc.de>
    // and Jamie Lokier <nospam@cern.ch> for this nice idea.

    // Open the temp file.
    fdi = open(tmpname, O_RDONLY, 0);
    if (fdi < 0)
        goto error;

    // Set the file close-on-exec. Ignore errors.
    {
        int flags;
        flags = fcntl(fdi, F_GETFD, 0);
        if (flags >= 0)
            (void) fcntl(fdi, F_SETFD, flags | FD_CLOEXEC);
    }

    // Now it's safe to unlink the temp file as it is still open.
    unlink(tmpname);

    // Compute name of temp fdi.
    {
        char *p = tmpfd + 14;
        {
            unsigned k = fdi;
            do {
                p++;
                k /= 10;
            } while (k > 0);
        }
        *p = 0;
        {
            unsigned k = fdi;
            do {
                *--p += k % 10;
                k /= 10;
            } while (k > 0);
        }
    }

    // Execute the original program via /proc/self/fd/X.
    execve(tmpfd, argv, envp);

    // The execve failed. Try again with /dev/fd/X as a fallback.
#if defined(__i386__)
    * ((uint32_t *) (tmpfd + 6)) = 0x7665642f;  // "/dev"
#else
    tmpfd[6] = '/'; tmpfd[7] = 'd'; tmpfd[8] = 'e'; tmpfd[9] = 'v';
#endif
    execve(tmpfd + 6, argv, envp);

#endif /* USE_FORK */


    //
    // ----- Step 7: error exit -----
    //

    // If we return from execve() there was an error. Give up.

error:
    (void) unlink(tmpname);
error1:
    // Note: the kernel will close all open files and
    //       unmap any allocated memory.
    (void) exit(127);
}


/*
vi:ts=8:et:nowrap
*/

